# -*- coding: utf-8 -*-
"""
Module to build and save dry (land) cost raster layers
"""
import logging
from pathlib import Path

import numpy as np
import numpy.typing as npt
from typing import Optional, Dict, Union, List, Tuple

from reVX.handlers.layered_h5 import LayeredTransmissionH5
from reVX.least_cost_xmission.config import (XmissionConfig, LandUseClasses,
                                             LandUseMultipliers,
                                             SlopeMultipliers, IsoMultipliers)
from reVX.least_cost_xmission.config.constants import (CELL_SIZE,
                                                       WATER_MULT,
                                                       WATER_NLCD_CODE,
                                                       METERS_IN_MILE,
                                                       HILL_MULT, MTN_MULT,
                                                       HILL_SLOPE, MTN_SLOPE,
                                                       DEFAULT_DTYPE,
                                                       DRY_MULTIPLIER_TIFF)
from reVX.least_cost_xmission.layers.masks import MaskArr
from reVX.least_cost_xmission.layers.base import BaseLayerCreator

logger = logging.getLogger(__name__)


NLCD_LAND_USE_CLASSES: LandUseClasses = {
    'cropland': [80, 81],
    'forest': [41, 42, 43],
    'wetland': [90, 95],
    'suburban': [21, 22, 23],
    'urban': [24],
}


class DryCostCreator(BaseLayerCreator):
    """
    Class to create and save dry transmission cost layers
    """
    def __init__(self, io_handler: LayeredTransmissionH5,
                 mask, output_tiff_dir=".",
                 dtype: npt.DTypeLike = DEFAULT_DTYPE,
                 cell_size=CELL_SIZE):
        """
        Parameters
        ----------
        io_handler : :class:`LayeredTransmissionH5`
            Transmission layer IO handler
        mask : ndarray
            Array representing mask for layer values.
        output_tiff_dir : path-like, optional
            Directory where cost layers should be saved as GeoTIFF.
            By default, ``"."``.
        dtype : np.dtype, optional
            Data type for final dataset. By default, ``float32``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.
        """
        super().__init__(io_handler=io_handler, mask=mask,
                         output_tiff_dir=output_tiff_dir, dtype=dtype,
                         cell_size=cell_size)
        self._iso_lookup: Dict[str, int] = {}

    def build(self, iso_region_tiff: str, nlcd_tiff: str, slope_tiff: str,
              cost_configs: Optional[Union[str, Dict]] = None,
              default_mults: Optional[IsoMultipliers] = None,
              extra_tiffs: Optional[List[str]] = None):
        """
        Build cost rasters using base line costs and multipliers. Save to
        GeoTIFF.

        Parameters
        ----------
        iso_region_tiff : str
            Path to the ISO region GeoTIFF
        nlcd_tiff : str
            Path to the National Land Coverage Database GeoTIFF
        slope_tiff : str
            Path to the slope GeoTIFF. Slope is in decimal percent.
        cost_configs : dict | str, optional
            Dictionary or path to json file contianing dictionary with
            Xmission cost configuration values. Valid configuration keys
            are:

                - "base_line_costs"
                - "iso_lookup"
                - "iso_multipliers"
                - "land_use_classes"
                - "new_substation_costs"
                - "power_classes"
                - "power_to_voltage"
                - "transformer_costs"
                - "upgrade_substation_costs"

            Each of these keys should point to a dictionary or a path to
            a separate json file contianing a dictionary of
            configurations for each section.
        default_mults : IsoMultipliers, optional
            Multipliers for regions not specified in iso_mults_fpath.
            by default None
        extra_tiffs : list, optional
            Optional list of extra GeoTIFFs to add to cost H5 file (e.g.
            a transmission barrier file). By default, ``None``, which
            does not add any extra layers.
        """
        xc = XmissionConfig(config=cost_configs)
        self._iso_lookup = xc['iso_lookup']

        logger.debug('Loading ISO region, slope and land use rasters')
        iso_layer = self._io_handler.load_data_using_h5_profile(
            iso_region_tiff, reproject=True)
        slope_layer = self._io_handler.load_data_using_h5_profile(
            slope_tiff, reproject=True)
        nlcd_layer = self._io_handler.load_data_using_h5_profile(
            nlcd_tiff, reproject=True)
        logger.debug('Loading complete')

        lu_classes = xc['land_use_classes']
        iso_mults: List[IsoMultipliers] = xc['iso_multipliers']
        mults_arr = self._compute_multipliers(iso_mults, iso_layer,
                                              slope_layer, nlcd_layer,
                                              land_use_classes=lu_classes,
                                              default_mults=default_mults)

        logger.debug('Saving multipliers array GeoTIFF')
        mult_tiff = self.output_tiff_dir / DRY_MULTIPLIER_TIFF
        self._io_handler.save_data_using_h5_profile(mults_arr, mult_tiff)

        if self._io_handler is not None:
            tiff_layers = [iso_region_tiff, slope_tiff, nlcd_tiff, mult_tiff]
            tiff_layers += extra_tiffs or []
            for layer_fp in tiff_layers:
                layer_name = Path(layer_fp).stem
                out = self._io_handler.load_data_using_h5_profile(
                    layer_fp, reproject=True)
                logger.debug(f'Writing {layer_name} to H5')
                self._io_handler.write_layer_to_h5(out, layer_name)

        for power_class, capacity in xc['power_classes'].items():
            logger.info('Calculating costs for class %s using a %sMW line',
                        power_class, capacity)
            blc_arr = self._compute_base_line_costs(capacity,
                                                    xc['base_line_costs'],
                                                    iso_layer)

            base_costs_tiff = 'base_line_costs_{}MW.tif'.format(capacity)
            out_fp = self.output_tiff_dir / base_costs_tiff
            self._io_handler.save_data_using_h5_profile(blc_arr, out_fp)

            # Calculate total costs w/ multipliers
            costs_arr = blc_arr * mults_arr

            dry_layer_name = 'tie_line_costs_{}MW'.format(capacity)
            tie_line_costs_tiff = '{}.tif'.format(dry_layer_name)
            out_fp = self.output_tiff_dir / tie_line_costs_tiff
            costs_arr[~self._mask] = 0
            self._io_handler.save_data_using_h5_profile(costs_arr, out_fp)
            if self._io_handler is not None:
                out = self._io_handler.load_data_using_h5_profile(
                    out_fp, reproject=True)
                logger.debug('Writing dry costs to H5')
                self._io_handler.write_layer_to_h5(out, dry_layer_name)

    @staticmethod
    def _compute_slope_mult(slope: npt.NDArray,
                            config: Optional[SlopeMultipliers] = None
                            ) -> npt.NDArray:
        """
        Create slope multiplier raster for a region. Unspecified slopes are
        left at 1.0

        Parameters
        ----------
        slope : numpy.ndarray
            Slope raster clipped to a region- "Terrain slope in % of grade"
        config : dict | None
            Multipliers and slope cut offs for hilly and mountain terrain.
            Use defaults if None.
            config['hill_mult'] : float
                Multiplier for hilly terrain
            config['mtn_slope'] : float
                Multiplier for mountainous terrain
            config['hill_slope'] : float
                Slope at and above which a cell is classified as hilly.
            config['mtn_slope'] : float
                Slope at and above which a cell is classified as mountainous

        Returns
        -------
        numpy.ndarray
            Slope multiplier raster. Minimum value for any cell is 1.
        """
        if config is None:
            config = {}

        hill_mult = config.get('hill_mult', HILL_MULT)
        mtn_mult = config.get('mtn_mult', MTN_MULT)
        hill_slope = config.get('hill_slope', HILL_SLOPE)
        mtn_slope = config.get('mtn_slope', MTN_SLOPE)

        hilly = (slope >= hill_slope) & (slope < mtn_slope)
        mountainous = slope >= mtn_slope

        mult_raster = np.ones(slope.shape, dtype=np.float32)
        mult_raster[hilly] = hill_mult
        mult_raster[mountainous] = mtn_mult

        return mult_raster

    @staticmethod
    def _compute_land_use_mult(land_use: npt.NDArray,
                               multipliers: LandUseMultipliers,
                               land_use_classes: LandUseClasses
                               ) -> npt.NDArray:
        """
        Convert NLCD raster to land use multiplier raster for a single ISO
        region. Land classes without specified multipliers are left a 1.

        Parameters
        ----------
        land_use : numpy.ndarray
           NLCD land user raster clipped to a region
        multipliers : LandUseMultipliers
            Multiplier for for land classes, E.g. {'forest': 1.5}
        land_use_classes : dict
            NCLD land use codes corresponding to use classes for multipliers

        Returns
        -------
        numpy.ndarray
            Land use multiplier raster. Minimum value for any cell is 1.
        """
        mult_raster = np.ones(land_use.shape, dtype=np.float32)

        # Determine mask arrays for NCLD values and multiplier to apply
        indices: List[Tuple[MaskArr, float, int]] = []
        multiplier: float
        for _class, multiplier in multipliers.items():  # type: ignore
            if _class not in land_use_classes:
                msg = ('Class {} not in land_use_classes: {}'
                       .format(_class, land_use_classes))
                logger.error(msg)
                raise ValueError(msg)

            nlcd_values: List[int] = land_use_classes[_class]  # type: ignore
            assert isinstance(nlcd_values, list)

            for nlcd_value in nlcd_values:
                index: MaskArr = land_use == nlcd_value
                indices.append((index, multiplier, nlcd_value))

        # Apply multipliers to appropriate cells
        for i in indices:
            mult_raster[i[0]] = i[1]

        return mult_raster

    def _compute_multipliers(self, iso_mults: List[IsoMultipliers],
                             iso_layer: npt.NDArray,
                             slope_layer: npt.NDArray,
                             land_use_layer: npt.NDArray,
                             land_use_classes: Optional[LandUseClasses] = None,
                             default_mults: Optional[IsoMultipliers] = None
                             ) -> npt.NDArray:
        """
        Create costs multiplier raster

        Parameters
        ----------
        iso_mults : List[IsoMultipliers]
            Land use and slope multipliers for ISOs
        sloper_layer : str, optional
            Name of slope layer in excl_h5, by default 'srtm_slope'
        ncld_layer : str, optional
            Name of NLCD (land use) layer in excl_h5, by default
            'usa_mrlc_nlcd2011'
        land_use_classes : dict, optional
            NCLD land use codes corresponding to use classes for multipliers.
            If None, use NLCD_LAND_USE_CLASSES. By default None
        default_mults : dict, optional
            Multipliers for regions not specified in iso_mults_fpath.
            By default None

        Returns
        -------
        numpy.ndarray
            Costs multiplier raster, including slope and land use
        """
        mults_arr = np.ones(self.shape, dtype=np.float32)
        regions_mask: MaskArr = np.full(self.shape, False, dtype=bool)

        if land_use_classes is None:
            land_use_classes = NLCD_LAND_USE_CLASSES

        for r_conf in iso_mults:
            iso_name = r_conf['iso']
            logger.info('Processing multipliers for region %s', iso_name)

            iso = self._iso_lookup[iso_name]
            logger.debug('ISO %s has id %s', iso_name, iso)

            mask: MaskArr = iso_layer == iso
            regions_mask = regions_mask | mask

            if 'land_use' in r_conf:
                r_lu = land_use_layer[mask]
                lu_mult = self._compute_land_use_mult(r_lu, r_conf['land_use'],
                                                      land_use_classes)
                mults_arr[mask] = lu_mult

            if 'slope' in r_conf:
                r_slope = slope_layer[mask]
                slope_mult = self._compute_slope_mult(r_slope, r_conf['slope'])
                mults_arr[mask] = mults_arr[mask] * slope_mult

        # Calculate multipliers for regions not defined in `config`
        logger.debug('Processing default region')
        if default_mults is not None:
            default_mask = ~regions_mask

            if 'land_use' in default_mults:
                region_land_use = land_use_layer[default_mask]
                lu_mult_dict = default_mults['land_use']
                lu_mult = self._compute_land_use_mult(region_land_use,
                                                      lu_mult_dict,
                                                      land_use_classes)
                mults_arr[default_mask] = lu_mult

            if 'slope' in default_mults:
                region_slope = slope_layer[default_mask]
                slope_mult = self._compute_slope_mult(region_slope,
                                                      default_mults['slope'])
                mults_arr[default_mask] = mults_arr[default_mask] * slope_mult

        # Set water multiplier last so we don't get super high multipliers at
        # water body boundaries next to steep slopes
        mults_arr[land_use_layer == WATER_NLCD_CODE] = WATER_MULT

        return mults_arr

    def _compute_base_line_costs(self, capacity: int,
                                 base_line_costs: Dict[str, Dict[str, float]],
                                 iso_layer: npt.NDArray) -> npt.NDArray:
        """
        Get base line cost per cell raster for a given voltage

        Parameters
        ----------
        capacity : int
            Desired line capacity (MW). This must be in base_line_costs.
        base_line_costs : dict
            Base line cost per mile for ISOs
        iso_layer: numpy.ndarray
            Raster defining ISO regions.

        Returns
        -------
        base_cost : numpy.ndarray
            Cost per cell raster in standard shape
        """
        base_cost = np.zeros(self.shape, dtype=np.float32)

        for iso in base_line_costs:
            logger.info('Processing costs for %s for %sMW', iso, capacity)
            iso_code = self._iso_lookup[iso]
            cost_per_mile = base_line_costs[iso][str(capacity)]
            cost_per_cell = cost_per_mile / METERS_IN_MILE * self._cell_size

            logger.debug('Base line $/mile is %s, $/cell is %s', cost_per_mile,
                         cost_per_cell)
            mask = iso_layer == iso_code
            base_cost[mask] = cost_per_cell

        return base_cost



# -*- coding: utf-8 -*-
"""
Abstract base calss for layer builders
"""
from pathlib import Path
from abc import ABC, abstractmethod

import numpy.typing as npt

from reVX.handlers.layered_h5 import LayeredTransmissionH5
from reVX.least_cost_xmission.config.constants import DEFAULT_DTYPE, CELL_SIZE


class BaseLayerCreator(ABC):
    """
    Abstract Base Class to create and save transmission routing layers
    """
    def __init__(self, io_handler: LayeredTransmissionH5,
                 mask=None, output_tiff_dir=".",
                 dtype: npt.DTypeLike = DEFAULT_DTYPE,
                 cell_size=CELL_SIZE):
        """
        Parameters
        ----------
        io_handler : :class:`LayeredTransmissionH5`
            Transmission layer IO handler
        mask : ndarray, optional
            Array representing mask for layer values. Only optional if
            subclass implementation handles masks differently
            (e.g. the `LayerCreator` class). By default, ``None``
        output_tiff_dir : path-like, optional
            Directory where cost layers should be saved as GeoTIFF.
            By default, ``"."``.
        dtype : np.dtype, optional
            Data type for final dataset. By default, ``float32``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.
        """
        self._io_handler = io_handler
        self.output_tiff_dir = Path(output_tiff_dir)
        self._mask = mask
        self._dtype = dtype
        self._cell_size = cell_size

    @property
    def shape(self):
        """tuple: Layer shape. """
        return self._io_handler.shape

    @abstractmethod
    def build(self, *args, **kwargs):
        """Build layer"""
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Module to create wind and solar plants for PLEXOS buses
"""
import json
import logging
import numpy as np
import os
import pandas as pd
from sklearn.metrics.pairwise import haversine_distances
from warnings import warn

from rex.resource import Resource
from rex.utilities import parse_table, SpawnProcessPool
from rex.utilities.utilities import to_records_array

from reVX.handlers.outputs import Outputs
from reVX.handlers.sc_points import SupplyCurvePoints
from reVX.utilities.utilities import log_versions
from reVX.plexos.base import BaseProfileAggregation, TZ_ALIASES

logger = logging.getLogger(__name__)


class Plants:
    """
    Base class to handle plants
    """

    def __init__(self, plants):
        self._plants = {}

        if isinstance(plants, (np.ndarray, list, tuple)):
            for i, plant in enumerate(plants):
                self[i] = plant
        elif isinstance(plants, dict):
            self._plants = plants

        self._i = 0

    def __repr__(self):
        msg = "{} with {} plants".format(self.__class__.__name__, len(self))
        return msg

    def __len__(self):
        return len(self._plants)

    def __getitem__(self, plant_id):
        """
        Get the plant build out for the given plant_id

        Parameters
        ----------
        plant_id : int | str
            Unique Id for plant of interest

        Returns
        -------
        list
            List of sc_gids' and thier associated resource gids to build plant
            from
        """
        return self._plants.get(plant_id, [])

    def __setitem__(self, plant_id, plant_build):
        """
        Update plant build

        Parameters
        ----------
        plant_id : int | str
            Unique Id for plant to update
        plant_build : list
            List of sc_gids' and thier associated resource gids to build plant
            from
        """
        self._plants[plant_id] = plant_build

    def __iter__(self):
        return self

    def __next__(self):
        if self._i >= len(self._plants):
            self._i = 0
            raise StopIteration

        plant_id = self.plant_ids[self._i]
        self._i += 1

        return self[plant_id]

    @property
    def plant_ids(self):
        """
        Plant ids

        Returns
        -------
        list
        """
        return list(self._plants.keys())

    @property
    def plants(self):
        """
        Dictionary matching plants to plant ids

        Returns
        -------
        dict
        """
        return self._plants

    @property
    def plant_builds(self):
        """
        List of plant builds

        Returns
        -------
        dict
        """
        plant_builds = {pid: pd.concat(plant, axis=1).T
                        for pid, plant in self.plants.items()
                        if plant is not None}

        return plant_builds

    @staticmethod
    def _parse_lists(column):
        """
        Check to see if list values are strings, if so parse with json.loads

        Parameters
        ----------
        column : pandas.Series
            Pandas DataFrame column to check

        Returns
        -------
        column : pandas.Series
            Pandas DataFrame column with values converted to lists if needed
        """
        if isinstance(column.iloc[0], str):
            column = column.apply(json.loads).values

        return column

    @classmethod
    def load(cls, plants_fpath):
        """
        Load pre-filled plants from disc

        Parameters
        ----------
        plants_fpath : str | DataFrame
            DataFrame or path to .csv containing pre-filled plants

        Returns
        -------
        Plants
            Initialized Plants instance with pre-filled plants
        """
        plant_builds = parse_table(plants_fpath)
        if 'plant_id' in plant_builds:
            plant_builds = plant_builds.set_index('plant_id')

        plant_builds = plant_builds.apply(cls._parse_lists)

        plants = {}
        for pid, build in plant_builds.iterrows():
            plant = []
            for i in range(len(build['sc_gids'])):
                sc_point = \
                    pd.Series({'sc_gid': build['sc_gids'][i],
                               'res_gids': build['res_gids'][i],
                               'gid_counts': build['gid_counts'][i],
                               'cf_means': build['res_cf_means'][i],
                               'build_capacity': build['build_capacity'][i]})
                plant.append(sc_point)

            plants[pid] = plant

        return cls(plants)


class PlexosPlants(Plants):
    """
    Class to identify and fill Plants
    """

    def __init__(self, plexos_table, sc_table, mymean_fpath,
                 dist_percentile=90, dist_thresh_km=None,
                 lcoe_col='total_lcoe', lcoe_thresh=1.3, offshore=False,
                 max_workers=None, plants_per_worker=40,
                 points_per_worker=400):
        """
        Parameters
        ----------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        mymean_fpath : str
            Path to reV multi-year-mean .h5 file (preferred) or annual reV
            generation output .h5 file. If annual, the plant buildouts might
            change from year to year (bad!).
        offshore : bool, optional
            Include offshore points, by default False
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        dist_thresh_km : float | None
            Optional absolute distance threshold in km that will override the
            dist_percentile input.
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        offshore : bool, optional
            Include offshore points, by default False
        max_workers : int, optional
            Number of workers to use for plant identification, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        """
        log_versions(logger)
        logger.info('Initializing PlexosPlants from plexos table with '
                    '{} rows'.format(len(plexos_table)))
        self._plant_table = self._parse_plant_table(plexos_table)
        self._capacity = self.plant_table['plant_capacity'].values.copy()
        self._cap_to_alloc = self.plant_table['plant_capacity'].values.copy()

        if max_workers is None:
            max_workers = os.cpu_count()

        self._sc_points = \
            SupplyCurvePoints(sc_table, mymean_fpath,
                              max_workers=max_workers,
                              points_per_worker=points_per_worker,
                              offshore=offshore)

        plants = self._identify_plants(self.plant_table,
                                       self._sc_points.sc_table,
                                       dist_percentile=dist_percentile,
                                       dist_thresh_km=dist_thresh_km,
                                       lcoe_col=lcoe_col,
                                       lcoe_thresh=lcoe_thresh,
                                       max_workers=max_workers,
                                       plants_per_worker=plants_per_worker)

        self._plants = {}
        self._fill_plants(plants)

        self._i = 0

    @property
    def plant_table(self):
        """
        Plants meta data table

        Returns
        -------
        pandas.DataFrame
        """
        return self._plant_table

    @property
    def plant_capacity(self):
        """
        Plant capacities in MW

        Returns
        -------
        ndarray
        """
        return self._capacity

    @property
    def sc_points(self):
        """
        Supply Curve Points

        Returns
        -------
        SupplyCurvePoints
        """
        return self._sc_points

    @staticmethod
    def _parse_plant_table(plexos_table):
        """
        Create Table of unique PLEXOS plants from plexos table

        Parameters
        ----------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table

        Returns
        -------
        plant_table : pandas.DataFrame
            Table of unique plants from plexos table
        """
        plexos_table = parse_table(plexos_table)
        if 'plant_id' not in plexos_table:
            plexos_table = \
                PlantProfileAggregation._parse_plexos_table(plexos_table)

        plant_table = \
            plexos_table.drop_duplicates('plant_id').sort_values('plant_id')

        cols = ['plant_id', 'plant_capacity', 'generator', 'busid', 'busname',
                'capacity', 'latitude', 'longitude', 'system']
        cols = [c for c in cols if c in plant_table]
        plant_table = plant_table[cols].set_index('plant_id')

        return plant_table

    @staticmethod
    def _check_coords(coords):
        """
        Check coordinate dimensions and units

        Parameters
        ----------
        coords : ndarray
            Either a single set or an array of (lat, lon) coordinates

        Returns
        -------
        coords : ndarray
            Coordinates in radians
        """
        if len(coords.shape) == 1:
            coords = np.expand_dims(coords, axis=0)

        if np.max(coords) > np.pi or np.min(coords) < - np.pi:
            coords = np.radians(coords)

        return coords

    @classmethod
    def _haversine_dist(cls, plant_coords, sc_coords):
        """
        Compute the haversine distance between the given plant(s) and given
        supply curve points

        Parameters
        ----------
        plant_coords : ndarray
            (lat, lon) coordinates of plant(s)
        sc_coords : ndarray
            n x 2 array of supply curve (lat, lon) coordinates

        Returns
        -------
        dist : ndarray
            Vector of distances between plant and supply curve points in km
        """
        plant_coords = cls._check_coords(plant_coords)
        sc_coords = cls._check_coords(sc_coords)

        dist = haversine_distances(plant_coords, sc_coords)
        if plant_coords.shape[0] == 1:
            dist = dist.flatten()

        R = 6371.0  # radius of the earth in kilometers

        return dist * R

    @staticmethod
    def _substation_distance(sc_table, percentile=90):
        """
        Determine the nth percentile of distance between substations and
        transmission from supply curve table

        Parameters
        ----------
        sc_table : pandas.DataFrame
            Supply curve table
        percentile : int, optional
            Percentile to compute substation to transmission distance for,
            by default 90

        Returns
        -------
        dist
            Nth percentile of distance between substations and transmission in
            km, used as plant search distance threshold
        """
        substations = sc_table['trans_type'] == "Substation"

        if 'dist_km' in sc_table:
            dist = sc_table.loc[substations, 'dist_km'].values
        elif 'dist_mi' in sc_table:
            dist = sc_table.loc[substations, 'dist_mi'].values * 1.6
        else:
            msg = ('Could not find "dist_mi" or "dist_km" in sc table '
                   'columns: {}'.format(sc_table.columns.values))
            logger.error(msg)
            raise KeyError(msg)

        return np.percentile(dist, percentile)

    @classmethod
    def _get_plant_sc_dists(cls, bus_coords, sc_table,
                            dist_percentile=90, dist_thresh_km=None,
                            lcoe_col='total_lcoe', lcoe_thresh=1.3):
        """
        Extract Supply curve gids and distances for plant originating at
        PLEXOS bus coords

        Parameters
        ----------
        bus_coords : ndarray
            bus (lat, lon) coordinates
        sc_table : pandas.DataFrame
            Supply Curve Table
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        dist_thresh_km : float | None
            Optional absolute distance threshold in km that will override the
            dist_percentile input.
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3

        Returns
        -------
        plant_sc : pandas.DataFrame
            Supply Curve for plant with distance to each sc_gid appended
        """
        logger.debug("Extracting supply curve gids for bus at {}"
                     .format(bus_coords))
        sc_coords = np.radians(sc_table[['latitude', 'longitude']].values)

        # Filter SC table to points within 'dist_tresh' of coords
        dist_km = cls._haversine_dist(bus_coords, sc_coords)

        if dist_thresh_km is None:
            dist_thresh_km = cls._substation_distance(
                sc_table, percentile=dist_percentile)

        logger.debug("- Using distance threshold of {} km"
                     .format(dist_thresh_km))
        while True:
            mask = dist_km <= dist_thresh_km
            plant_sc = sc_table[['latitude', 'longitude', lcoe_col]].copy()
            plant_sc = plant_sc.loc[mask]
            if len(plant_sc) > 1:
                break
            else:
                dist_thresh_km *= 1.2

        # Find lowest lcoe site
        pos = np.argmin(plant_sc[lcoe_col])
        lcoe_thresh = plant_sc.iloc[pos][lcoe_col] * lcoe_thresh
        plant_coords = \
            plant_sc.iloc[pos][['latitude', 'longitude']].values.astype(float)
        logger.debug("- Plant will be centered at {}".format(plant_coords))
        logger.debug("- Only supply curve points with an lcoe < {} will be "
                     "used".format(lcoe_thresh))

        # Filter SC table to lcoe values within 'lcoe_thresh' of min LCOE value
        sc_cols = ['sc_gid', lcoe_col]
        plant_sc = sc_table[sc_cols].copy()
        plant_sc["bus_dist"] = dist_km
        mask = plant_sc[lcoe_col] <= lcoe_thresh
        plant_sc = plant_sc.loc[mask]
        sc_coords = sc_coords[mask]

        # Sort by distance
        plant_sc['dist'] = cls._haversine_dist(plant_coords, sc_coords)
        plant_sc = plant_sc.sort_values('dist')

        return plant_sc.reset_index(drop=True)

    @classmethod
    def _identify_plants(cls, plant_table, sc_table,
                         dist_percentile=90, dist_thresh_km=None,
                         lcoe_col='total_lcoe', lcoe_thresh=1.3,
                         max_workers=None, plants_per_worker=40):
        """
        Identify plant associated with each bus and return supply curve table

        Parameters
        ----------
        plant_table : pandas.DataFrame
            Table of unique plants from plexos table
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        dist_thresh_km : float | None
            Optional absolute distance threshold in km that will override the
            dist_percentile input.
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        max_workers : int, optional
            Number of workers to use for plant identification, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40

        Returns
        -------
        plants : list
            List of supply curve points that can be used to fill each plant of
            interest
        """
        plants = []
        if max_workers is None:
            max_workers = os.cpu_count()

        if max_workers > 1:
            logger.info('Identifying plants in parallel')
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                slices = SupplyCurvePoints._create_worker_slices(
                    plant_table, points_per_worker=plants_per_worker)
                for table_slice in slices:
                    future = exe.submit(cls._identify_plants,
                                        plant_table.iloc[table_slice].copy(),
                                        sc_table,
                                        dist_percentile=dist_percentile,
                                        dist_thresh_km=dist_thresh_km,
                                        lcoe_col=lcoe_col,
                                        lcoe_thresh=lcoe_thresh,
                                        max_workers=1)
                    futures.append(future)

                for i, future in enumerate(futures):
                    plants.extend(future.result())
                    logger.debug('Identified {} out of {} plants'
                                 .format((i + 1) * plants_per_worker,
                                         len(plant_table)))
        else:
            logger.info('Identifying plants in serial')
            for i, bus in plant_table.iterrows():
                coords = \
                    bus[['latitude', 'longitude']].values.astype(float)
                plant = cls._get_plant_sc_dists(
                    coords, sc_table,
                    dist_percentile=dist_percentile,
                    dist_thresh_km=dist_thresh_km,
                    lcoe_col=lcoe_col,
                    lcoe_thresh=lcoe_thresh)
                plants.append(plant)
                logger.debug('Identified {} out of {} PlexosPlants.'
                             .format(i + 1, len(plant_table)))

        return plants

    @staticmethod
    def _get_sc_gids(identified_plants, idx):
        """
        For all plants extract sc_gid, dist (to sc_gid), and bus_dist from
        Supply Curve points

        Parameters
        ----------
        identified_plants : list
            List of identified plants, I.E., Supply curve points available
            to fill each plant along with the distance to the plant center
            and associated bus.
        idx : int
            index to extract from plant sc_tables

        Returns
        -------
        tuple
            (sc_gids, dists, bus_dists)
        """
        sc_gids = []
        bus_dists = []
        dists = []
        for plant in identified_plants:
            sc_point = plant.iloc[idx]
            sc_gids.append(sc_point['sc_gid'])
            bus_dists.append(sc_point['bus_dist'])
            dists.append(sc_point['dist'])

        bus_dists = np.array(bus_dists)
        dists = np.array(dists)

        return sc_gids, dists, bus_dists

    def _allocate_sc_gids(self, sc_gids, dists, bus_dists):
        """
        Allocate capacity (MW) from supply curve points to plants

        Parameters
        ----------
        sc_gids : list
            List of supply curve point gids to allocate capacity from
        dists : list
            List of distances from plants to sc_gids
        bus_dists : list
            List of distances from bus associated with plants to sc_gids
        """

        unique_gids, plant_gids = np.unique(sc_gids, return_inverse=True)
        for i, sc_gid in enumerate(unique_gids):
            sc_gid = int(sc_gid)
            if self.sc_points.check_sc_gid(sc_gid):
                plant_ids = np.where(plant_gids == i)[0]
                if len(plant_ids) > 1:
                    sc_dists = dists[plant_ids]
                    if len(sc_dists) != len(np.unique(sc_dists)):
                        idxs = np.argsort(bus_dists[plant_ids])
                    else:
                        idxs = np.argsort(sc_dists)

                    plant_ids = plant_ids[idxs]

                for plant_id in plant_ids:
                    capacity = self._cap_to_alloc[plant_id]
                    if (capacity > 0) and self.sc_points.check_sc_gid(sc_gid):
                        built_point = self.sc_points.get_capacity(sc_gid,
                                                                  capacity)
                        if built_point is not None:
                            built_cap = built_point['build_capacity']
                            plant = self[plant_id]
                            plant.append(built_point)

                            self[plant_id] = plant
                            self._cap_to_alloc[plant_id] -= built_cap
                            logger.debug('Allocating {:.1f}MW to plant {} from'
                                         ' sc_gid {}'.format(built_cap,
                                                             plant_id,
                                                             sc_gid))
                        else:
                            msg = ('WARNING: sc_gid {} returned 0 capacity!'
                                   .format(sc_gid))
                            logger.warning(msg)
                            warn(msg)

    def _fill_plants(self, identified_plants):
        """
        Fill plants with capacity (MW) from supply curve points

        Parameters
        ----------
        identified_plants : list
            List of identified plants, I.E., Supply curve points available
            to fill each plant along with the distance to the plant center
            and associated bus.
        """
        i = 0
        total_cap = np.sum(self.plant_capacity)
        while np.any(self._cap_to_alloc > 0):
            cap_remaining_0 = sum(self._cap_to_alloc)
            logger.info('Allocating sc_gids to plants round {}'
                        .format(i))
            sc_gids, dists, bus_dists = self._get_sc_gids(identified_plants, i)
            self._allocate_sc_gids(sc_gids, dists, bus_dists)
            cap_remaining_1 = sum(self._cap_to_alloc)
            i_cap_allocated = cap_remaining_0 - cap_remaining_1
            logger.info('{:.1f} MW allocated in round {}'
                        .format(i_cap_allocated, i))
            i += 1
            logger.info('{:.1f} MW allocated out of {:.1f} MW'
                        .format(total_cap - cap_remaining_1, total_cap))
            logger.info('{} of {} plants have been filled'
                        .format(np.sum(self._cap_to_alloc <= 0), len(self)))

    def dump(self, out_fpath=None):
        """
        Create plants meta data from filled plants DataFrames:
            - Location (lat, lon)
            - final capacity (MW)
            - sc_gids
            - res_gids
            - res gid_counts

        Parameters
        ----------
        out_fpath : str, optional
            .csv path to save plant meta data too, by default None

        Returns
        -------
        plants_meta : pandas.DataFrame
            Location (lat, lon), final capacity (MW), and associated sc_gids,
            res_gids, and res gid_counts for all plants
        """
        plants_meta = []
        for pid, plant in self.plant_builds.items():
            plants_meta.append(pd.Series(
                {'sc_gids': plant['sc_gid'].values.tolist(),
                 'res_gids': plant['res_gids'].values.tolist(),
                 'gid_counts': plant['gid_counts'].values.tolist(),
                 'res_cf_means': plant['cf_means'].values.tolist(),
                 'build_capacity': plant['build_capacity'].values.tolist()},
                name=pid))

        plants_meta = pd.concat(plants_meta, axis=1).T
        plants_meta.index.name = 'plant_id'

        if out_fpath:
            plants_meta.to_csv(out_fpath)

        return plants_meta

    @classmethod
    def save(cls, plexos_table, sc_table, mymean_fpath, out_fpath,
             dist_percentile=90, lcoe_col='total_lcoe', lcoe_thresh=1.3,
             offshore=False, max_workers=None, plants_per_worker=40,
             points_per_worker=400):
        """
        Identify, fill, and then save plants to disc

        Parameters
        ----------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        mymean_fpath : str
            Path to reV multi-year-mean .h5 file (preferred) or annual reV
            generation output .h5 file. If annual, the plant buildouts might
            change from year to year (bad!).
        out_fpath : str
            .csv path to save plant meta data too
        offshore : bool, optional
            Include offshore points, by default False
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        offshore : bool, optional
            Include offshore points, by default False
        max_workers : int, optional
            Number of workers to use for plant identification, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        """
        pp = cls(plexos_table, sc_table, mymean_fpath,
                 dist_percentile=dist_percentile, lcoe_col=lcoe_col,
                 lcoe_thresh=lcoe_thresh, offshore=offshore,
                 max_workers=max_workers, plants_per_worker=plants_per_worker,
                 points_per_worker=points_per_worker)

        pp.dump(out_fpath=out_fpath)


class PlantProfileAggregation:
    """
    Aggregate renewable generation profiles to Plexos "plants"
    """

    def __init__(self, plexos_table, sc_table, mymean_fpath, cf_fpath,
                 plants=None, dist_percentile=90, dist_thresh_km=None,
                 lcoe_col='total_lcoe', lcoe_thresh=1.3,
                 offshore=False, max_workers=None,
                 plants_per_worker=40, points_per_worker=400,
                 plant_name_col=None, tech_tag=None, dset_tag='',
                 timezone='UTC', bespoke=False):
        """
        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity (MW) provided as a .csv,
            .json, or pandas DataFrame. Needs columns: latitude, longitude,
            capacity. Note that capacity needs to be AC for wind and DC for
            solar.
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        mymean_fpath : str
            Path to reV multi-year-mean output .h5 file to pull cf_mean-means
            from - this will be used to determine where plants are built so it
            does not differ from year to year. The meta must match cf_fpath.
        cf_fpath : str
            Path to reV annual Generation output .h5 file to pull CF profiles
            from. Meta must match mymean_fpath.
        plants : PlexosPlants | None
            Optional PlexosPlants input. If None, PlexosPlants object will
            be Initialized from the plexos table input.
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        max_workers : int, optional
            Number of workers to use for point and plant creation, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        offshore : bool, optional
            Include offshore points, by default False
        plant_name_col : str | None
            Column in plexos_table that has the plant name that should be used
            in the plexos output csv column headers.
        tech_tag : str | None
            Optional technology tag to include as a suffix in the plexos output
            csv column headers.
        dset_tag : str
            Dataset tag to append to dataset names in cf profile file. e.g. If
            the cf profile file is a multi year file using dset_tag="-2008"
            will enable us to select the corresponding datasets
            (cf_mean-2008, cf_profile-2008, etc)
        timezone : str
            Timezone for output generation profiles. This is a string that will
            be passed to pytz.timezone() e.g. US/Pacific, US/Mountain,
            US/Central, US/Eastern, or UTC. For a list of all available
            timezones, see pytz.all_timezones
        bespoke : bool
            Flag to signify if the cf_fpath file was generated using the reV
            bespoke wind module. The bespoke output files have generation
            profiles at the supply curve grid resolution which is different
            than traditional reV generation outputs that are on the resource
            grid resolution.
        """
        log_versions(logger)
        logger.info('Initializing PlantProfileAggregation')

        self._plexos_table = self._parse_plexos_table(
            plexos_table, plant_name_col=plant_name_col)

        self._plant_name_col = plant_name_col
        self._tech_tag = tech_tag
        self._dset_tag = dset_tag
        self._timezone = timezone
        self._mymean_fpath = mymean_fpath
        self._cf_fpath = cf_fpath
        self._cf_gid_map = self._parse_cf_gid_map(cf_fpath)
        self._sc_table = SupplyCurvePoints._parse_sc_table(sc_table,
                                                           offshore=offshore)

        if bespoke:
            self._sc_table = BaseProfileAggregation.convert_bespoke_sc(
                self._sc_table, 'gid')

        if plants is None:
            self._plants = PlexosPlants(self._plexos_table, self._sc_table,
                                        mymean_fpath,
                                        dist_percentile=dist_percentile,
                                        dist_thresh_km=dist_thresh_km,
                                        lcoe_col=lcoe_col,
                                        lcoe_thresh=lcoe_thresh,
                                        offshore=offshore,
                                        max_workers=max_workers,
                                        plants_per_worker=plants_per_worker,
                                        points_per_worker=points_per_worker)
        else:
            self._plants = Plants.load(plants)

        self._sc_bus_dist = None

    def __repr__(self):
        msg = "{} with {} plants".format(self.__class__.__name__, len(self))
        return msg

    def __len__(self):
        return len(self.plexos_table)

    @property
    def plexos_table(self):
        """
        PLEXOS table

        Returns
        -------
        pandas.DataFrame
        """
        return self._plexos_table

    @property
    def mymean_fpath(self):
        """
        reV multi year mean output file path

        Returns
        -------
        str
        """
        return self._mymean_fpath

    @property
    def cf_fpath(self):
        """
        reV generation output file path

        Returns
        -------
        str
        """
        return self._cf_fpath

    @property
    def cf_gid_map(self):
        """Mapping of res_gid (index) to gen_gid (values)

        Returns
        -------
        pd.Series
        """
        return self._cf_gid_map

    @property
    def sc_table(self):
        """
        Supply Curve Table

        Returns
        -------
        pandas.DataFrame
        """
        return self._sc_table

    @property
    def plants(self):
        """
        Dictionary matching plants to plant ids

        Returns
        -------
        dict
        """
        return self._plants.plants

    @property
    def plant_builds(self):
        """
        PLEXOS Plant builds

        Returns
        -------
        dict
        """
        return self._plants.plant_builds

    @property
    def sc_bus_dist(self):
        """
        Compute distance between every Supply Curve gid and every PLEXOS bus

        Returns
        -------
        ndarray
        """
        if self._sc_bus_dist is None:
            cols = ['latitude', 'longitude']
            plant_coords = self._plexos_table[cols].values.astype(float)
            sc_coords = self.sc_table[cols].values.astype(float)

            self._sc_bus_dist = \
                PlexosPlants._haversine_dist(plant_coords, sc_coords).T

        return self._sc_bus_dist

    @property
    def unique_plant_names(self):
        """Get a list of ordered unique plant names if plant_name_col
        was provided.

        Returns
        -------
        list | None
        """
        names = None

        if self._plant_name_col is not None:
            names = BaseProfileAggregation.get_unique_plant_names(
                self.plexos_table, self._plant_name_col, self._tech_tag)

        return names

    @property
    def tz_alias(self):
        """Get a short 3-char tz alias if the timezone is common in the US
        (pst, mst, cst, est)

        Returns
        -------
        str
        """
        return TZ_ALIASES.get(self._timezone, self._timezone)

    def get_gen_gid(self, res_gid):
        """Get a generation gid from a resource gid using cf_gid_map. Accounts
        for a many-to-one resource-to-gen_gid mapping.

        Parameters
        ----------
        res_gid : int

        Returns
        -------
        gen_gid
        """

        res_gid = self.cf_gid_map.loc[res_gid]

        if isinstance(res_gid, pd.Series):
            res_gid = res_gid.values[0]

        return res_gid

    @staticmethod
    def _parse_plexos_table(plexos_table, plant_name_col=None):
        """
        Parse PLEXOS table from file and reduce to PLEXOS_COLS
        Combine buses at the same coordinates and add unique plant_ids

        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity (MW) provided as a .csv,
            .json, or pandas DataFrame
        plant_name_col : str | None
            Column in plexos_table that has the plant name that should be used
            in the plexos output csv column headers.

        Returns
        -------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table
        """
        logger.info('Parsing plexos table')
        plexos_table = parse_table(plexos_table)
        cols = ['generator', 'busid', 'busname', 'capacity', 'latitude',
                'longitude', 'system']

        rename = {c: c.lower() for c in plexos_table}
        plexos_table = plexos_table.rename(columns=rename)

        if plant_name_col is not None:
            msg = ('plant_name_col "{}" not in plexos table!'
                   .format(plant_name_col))
            assert plant_name_col.lower() in plexos_table, msg
            cols.append(plant_name_col.lower())

        cols = [c.lower() for c in plexos_table if c.lower() in cols]
        plexos_table = plexos_table[cols]

        mask = plexos_table['latitude'] > 90
        mask |= plexos_table['latitude'] < -90
        mask |= plexos_table['longitude'] > 180
        mask |= plexos_table['longitude'] < -180
        if np.any(mask):
            msg = ('WARNING: {} Buses have invalid coordinates:\n{}'
                   .format(np.sum(mask), plexos_table.loc[mask]))
            logger.warning(msg)
            warn(msg)
            plexos_table = plexos_table.loc[~mask]

        mask = plexos_table['capacity'] > 0
        plexos_table = plexos_table.loc[mask]
        cols = ['latitude', 'longitude']
        plant_cap = plexos_table.groupby(cols)['capacity'].sum()
        plant_cap = plant_cap.reset_index().reset_index()
        rename = {'index': 'plant_id', 'capacity': 'plant_capacity'}
        plant_cap = plant_cap.rename(columns=rename)

        plexos_table = plexos_table.merge(plant_cap,
                                          on=cols,
                                          how='inner')

        return plexos_table

    @staticmethod
    def _parse_cf_gid_map(cf_fpath):
        """
        Map resource gids to gen gids

        Parameters
        ----------
        cf_fpath : str
            Path to reV generation output .h5 file

        Returns
        -------
        cf_gid_map : pd.Series
            Mapping of res_gid (index) to gen_gid (values)
        """
        logger.info('Mapping reV resource GIDs to generation GIDs.')
        with Resource(cf_fpath) as f:
            res_gids = f.get_meta_arr('gid')

        if not isinstance(res_gids, np.ndarray):
            res_gids = np.array(list(res_gids))

        cf_gid_map = pd.Series(np.arange(len(res_gids)), index=res_gids,
                               name='gen_gid')
        cf_gid_map.index.name = 'res_gid'

        return cf_gid_map

    @staticmethod
    def _collapse_multi_list(multi_list):
        """Collapse a list of lists into one list"""
        if isinstance(multi_list, (pd.Series, pd.DataFrame)):
            multi_list = multi_list.values.tolist()

        if any(isinstance(x, (list, tuple)) for x in multi_list):
            multi_list = [item for sublist in multi_list for item in sublist]

        return multi_list

    def plants_meta(self):
        """Create plants meta data from filled plants DataFrames:

            - Location (lat, lon)
            - final capacity (MW)
            - sc_gids
            - res_gids
            - res gid_counts

        Parameters
        ----------
        plants : list
            List of filled plant DataFrames
        out_fpath : str
            .csv path to save plant meta data too

        Returns
        -------
        plants_meta : pandas.DataFrame
            Location (lat, lon), final capacity, and associated sc_gids,
            res_gids, and res gid_counts for all plants
        """
        plants_meta = []
        for pid, plant in self.plant_builds.items():
            single_meta = {
                'sc_gids': plant['sc_gid'].values.tolist(),
                'res_gids': plant['res_gids'].values.tolist(),
                'gid_counts': plant['gid_counts'].values.tolist(),
                'gen_gids': [[self.get_gen_gid(gid) for gid in gids]
                             for gids in plant['res_gids'].values],
                'res_cf_means': plant['cf_means'].values.tolist(),
                'build_capacity': plant['build_capacity'].values.tolist()}
            single_meta = {k: self._collapse_multi_list(v)
                           for k, v in single_meta.items()}
            single_meta['cf_mean'] = np.hstack(plant['cf_means'].values).mean()
            plants_meta.append(pd.Series(single_meta, name=pid))

        plants_meta = pd.concat(plants_meta, axis=1).T
        plants_meta.index.name = 'plant_id'

        plants_meta = self.plexos_table.merge(plants_meta.reset_index(),
                                              on='plant_id', how='outer')

        return plants_meta

    def _make_plant_meta(self, bus_meta):
        """
        Create plant meta data for given bus

        Parameters
        ----------
        bus_meta : pandas.Series
            Meta data for desired bus to build plant for

        Returns
        -------
        plant_meta : pandas.DataFrame
            Meta data for plant associated with given bus, constructed from:
            - Plant table
            - Supply Curve table
            - Bus capacity (MW)
        """
        plant_meta = self.plant_builds[bus_meta['plant_id']]
        plant_meta['gen_gids'] = \
            plant_meta['res_gids'].apply(lambda gids: [self.get_gen_gid(gid)
                                                       for gid in gids])

        sc_cols = ['res_gids', 'gen_gids', 'gid_counts', 'capacity']
        sc_cols = [c for c in self.sc_table if c not in sc_cols]
        plant_meta = plant_meta.merge(self.sc_table[sc_cols],
                                      on='sc_gid', how='left')

        plant_capacity = plant_meta['build_capacity'].sum()
        if plant_capacity != bus_meta['capacity']:
            bulid_capacity = (plant_meta['build_capacity'] / plant_capacity
                              * bus_meta['capacity'])
            plant_meta.loc[:, 'build_capacity'] = bulid_capacity

        return plant_meta

    @staticmethod
    def _make_profile(cf_fpath, plant_build, dset_tag=''):
        """
        Make generation profiles for given plant buildout

        Parameters
        ----------
        cf_fpath : str
            Path to reV Generation output .h5 file to pull CF profiles from
        plant_build : pandas.DataFrame
            DataFrame describing plant buildout:
                - Supply curve gids and the capacity (MW) to build at each
                    - res_gids, gen_gids, gid_counts by sc_gid
        dset_tag : str
            Dataset tag to append to dataset names in cf profile file. e.g. If
            the cf profile file is a multi year file using dset_tag="-2008"
            will enable us to select the corresponding datasets
            (cf_mean-2008, cf_profile-2008, etc)

        Returns
        -------
        profile: ndarray
            Generation profile for plant as a vector
        """
        with Resource(cf_fpath) as f:
            profile = None
            for _, row in plant_build.iterrows():
                gid_capacities = (row['gid_counts'] / np.sum(row['gid_counts'])
                                  * row['build_capacity'])
                cf_profiles = f['cf_profile' + dset_tag, :, row['gen_gids']]
                for i, cf_profile in enumerate(cf_profiles.T):
                    if profile is None:
                        profile = cf_profile * gid_capacities[i]
                    else:
                        profile += cf_profile * gid_capacities[i]

        if len(profile.shape) != 1:
            profile = profile.flatten()

        return profile

    def aggregate_profiles(self, out_fpath):
        """
        Aggregate plants from capacity factor profiles and save to given
        output .h5 path

        Parameters
        ----------
        out_fpath : str
            .h5 path to save aggregated plant profiles to
            A companion .csv with be saved at the same location for plexos.
        """

        if not out_fpath.endswith('.h5'):
            out_fpath = out_fpath + '.h5'

        if not os.path.exists(os.path.dirname(out_fpath)):
            os.makedirs(os.path.dirname(out_fpath))

        out_fpath = out_fpath.replace('.h5', f'_{self.tz_alias}.h5')

        with Outputs(out_fpath, mode='w') as f_out:
            f_out.set_version_attr()
            with Resource(self.cf_fpath) as f_in:
                logger.info('Copying time_index')
                ti = f_in['time_index' + self._dset_tag].tz_convert(None)
                f_out['time_index'] = ti

            logger.info('Writing meta data')
            f_out['meta'] = self.plants_meta()

            f_out.h5.create_group('plant_meta')
            gen_profiles = []
            logger.info('Extracting profiles and writing meta for plants')
            for i, irow in enumerate(self.plexos_table.iterrows()):
                bus_id, bus_meta = irow
                logger.debug('Building plant for bus {}'.format(bus_id))
                plant_meta = self._make_plant_meta(bus_meta)
                prof = self._make_profile(self.cf_fpath, plant_meta.copy(),
                                          dset_tag=self._dset_tag)
                gen_profiles.append(prof)

                plant_meta = to_records_array(plant_meta)
                logger.debug('Writing plant_meta/{} ({} out of {})'
                             .format(bus_id, i + 1, len(self.plexos_table)))
                f_out._create_dset('plant_meta/{}'.format(bus_id),
                                   plant_meta.shape,
                                   plant_meta.dtype,
                                   chunks=None,
                                   data=plant_meta)

            logger.info('Writing Generation Profiles')
            gen_profiles = np.dstack(gen_profiles)[0].astype('float32')
            gen_profiles = BaseProfileAggregation.tz_convert_profiles(
                gen_profiles, self._timezone)
            f_out._create_dset('gen_profiles',
                               gen_profiles.shape,
                               gen_profiles.dtype,
                               chunks=(None, 100),
                               data=gen_profiles,
                               attrs={'units': 'MW'})

        logger.info('Finished aggregating profiles to: {}'.format(out_fpath))

        df_plx = pd.DataFrame(gen_profiles, columns=self.unique_plant_names,
                              index=ti)
        df_plx.index.name = 'DATETIME'
        csv_fp = out_fpath.replace('.h5', '.csv')
        df_plx.to_csv(csv_fp)

        logger.info('Wrote plexos formatted profiles to: {}'.format(csv_fp))

    @classmethod
    def aggregate(cls, plexos_table, sc_table, mymean_fpath, cf_fpath,
                  plants_fpath, out_fpath, offshore=False):
        """
        Aggregate pre-filled plants

        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity (MW) provided as a .csv,
            .json, or pandas DataFrame
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        mymean_fpath : str
            Path to reV multi-year-mean output .h5 file to pull cf_mean-means
            from - this will be used to determine where plants are built so it
            does not differ from year to year. The meta must match cf_fpath.
        cf_fpath : str
            Path to reV annual Generation output .h5 file to pull CF profiles
            from. Meta must match mymean_fpath.
        plants_fpath : str
            Path to .csv containing pre-filled plants
        out_fpath : str
            .h5 path to save aggregated plant profiles to
        offshore : bool, optional
            Include offshore points, by default False
        """
        pp = cls(plexos_table, sc_table, mymean_fpath, cf_fpath,
                 plants=plants_fpath, offshore=offshore)
        # Add plants to PlexosPlant instance

        pp.aggregate_profiles(out_fpath)

    @classmethod
    def run(cls, plexos_table, sc_table, mymean_fpath, cf_fpath, out_fpath,
            dist_percentile=90, dist_thresh_km=None,
            lcoe_col='total_lcoe', lcoe_thresh=1.3,
            max_workers=None, points_per_worker=400, plants_per_worker=40,
            offshore=False, plant_name_col=None, tech_tag=None, dset_tag='',
            timezone='UTC', bespoke=False):
        """
        Find, fill, and save profiles for Plants associated with given PLEXOS
        buses

        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity (MW) provided as a .csv,
            .json, or pandas DataFrame. Needs columns: latitude, longitude,
            capacity. Note that capacity needs to be AC for wind and DC for
            solar.
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        mymean_fpath : str
            Path to reV multi-year-mean output .h5 file to pull cf_mean-means
            from - this will be used to determine where plants are built so it
            does not differ from year to year. The meta must match cf_fpath.
        cf_fpath : str
            Path to reV annual Generation output .h5 file to pull CF profiles
            from. Meta must match mymean_fpath.
        out_fpath : str
            .h5 path to save aggregated plant profiles to
            A companion .csv with be saved at the same location for plexos.
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        dist_thresh_km : float | None
            Optional absolute distance threshold in km that will override the
            dist_percentile input.
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        max_workers : int, optional
            Number of workers to use for point and plant creation, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        offshore : bool, optional
            Include offshore points, by default False
        plant_name_col : str | None
            Column in plexos_table that has the plant name that should be used
            in the plexos output csv column headers.
        tech_tag : str | None
            Optional technology tag to include as a suffix in the plexos output
            csv column headers.
        dset_tag : str
            Dataset tag to append to dataset names in cf profile file. e.g. If
            the cf profile file is a multi year file using dset_tag="-2008"
            will enable us to select the corresponding datasets
            (cf_mean-2008, cf_profile-2008, etc)
        timezone : str
            Timezone for output generation profiles. This is a string that will
            be passed to pytz.timezone() e.g. US/Pacific, US/Mountain,
            US/Central, US/Eastern, or UTC. For a list of all available
            timezones, see pytz.all_timezones
        bespoke : bool
            Flag to signify if the cf_fpath file was generated using the reV
            bespoke wind module. The bespoke output files have generation
            profiles at the supply curve grid resolution which is different
            than traditional reV generation outputs that are on the resource
            grid resolution.
        """
        pp = cls(plexos_table, sc_table, mymean_fpath, cf_fpath,
                 offshore=offshore,
                 dist_percentile=dist_percentile,
                 dist_thresh_km=dist_thresh_km,
                 lcoe_col=lcoe_col,
                 lcoe_thresh=lcoe_thresh,
                 max_workers=max_workers,
                 points_per_worker=points_per_worker,
                 plants_per_worker=plants_per_worker,
                 plant_name_col=plant_name_col,
                 tech_tag=tech_tag,
                 dset_tag=dset_tag,
                 timezone=timezone,
                 bespoke=bespoke)

        pp.aggregate_profiles(out_fpath)



"""
Least cost transmission utilities
"""
import logging
from typing import Optional, Union, Dict

import pandas as pd
import rasterio as rio
import geopandas as gpd
import numpy.typing as npt
from shapely.geometry import Point, LineString

from reVX.utilities import rasterize
from reVX.least_cost_xmission.config.constants import DEFAULT_DTYPE

VECTOR_CACHE: Dict[str, gpd.GeoDataFrame] = {}

logger = logging.getLogger(__name__)


def rasterize_shape_file(fname: str, profile: Dict,
                         buffer_dist: Optional[float] = None,
                         all_touched: bool = False,
                         reproject_vector: bool = True,
                         burn_value: Union[int, float] = 1,
                         boundary_only: bool = False,
                         dtype: npt.DTypeLike = DEFAULT_DTYPE,
                         ) -> npt.NDArray:
    """
    Rasterize a vector layer.

    Parameters
    ----------
    fname : str
        Full path to gpgk or shp file
    profile : dict
        Raster profile to use
    buffer_dist : float, optional
        Distance to buffer features in fname by. Same units as the
        template raster. By default, ``None``.
    all_touched : bool, optional
        Set all cells touched by vector to 1. False results in less cells
        being set to 1. By default, ``False``.
    reproject_vector : bool, optional
        Reproject CRS of vector to match template raster if True.
        By default, ``True``.
    burn_value : int | float, optional
        Value used to burn vectors into raster. By default, ``1``.
    boundary_only : bool, optional
        If True, rasterize boundary of vector. By default, ``False``.
    dtype : np.dtype, optional
        Datatype to use. By default, ``float32``.

    Returns
    -------
    numpy.nd_array
        Rasterized vector data
    """
    if fname in VECTOR_CACHE:
        logger.debug('Vector data for %s found in cache', fname)
        gdf = VECTOR_CACHE[fname]
    else:
        logger.debug('Loading %s', fname)
        gdf = gpd.read_file(fname)
        VECTOR_CACHE[fname] = gdf

    if reproject_vector:
        logger.debug('Reprojecting vector')
        gdf = gdf.to_crs(crs=profile['crs'])

    logger.debug('Rasterizing %s', fname)
    return rasterize(gdf, profile, buffer_dist=buffer_dist,
                     all_touched=all_touched, burn_value=burn_value,
                     boundary_only=boundary_only, dtype=dtype)


def convert_pois_to_lines(poi_csv_f: str, template_f: str, out_f: str):
    """
    Convert POIs in CSV to lines and save in a geopackage as substations. Also
    create a fake transmission line to connect to the substations.

    Parameters
    ----------
    poi_csv_f : str
        Path to CSV file with POIs in it
    template_f : str
        Path to template raster with CRS to use for geopackage
    out_f : str
        Path and file name for geopackage
    """
    logger.debug('Converting POIs in %s to lines in %s', poi_csv_f, out_f)
    with rio.open(template_f) as ras:
        crs = ras.crs

    df = pd.read_csv(poi_csv_f)[['POI Name', 'State', 'Voltage (kV)', 'Lat',
                                 'Long']]

    pts = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.Long, df.Lat))
    pts = pts.set_crs('EPSG:4326')
    pts = pts.to_crs(crs)

    # Convert points to short lines
    new_geom = []
    for pt in pts.geometry:
        end = Point(pt.x + 50, pt.y + 50)
        line = LineString([pt, end])
        new_geom.append(line)
    lines = pts.set_geometry(new_geom, crs=crs)

    # Append some fake values to make the LCP code happy
    lines['ac_cap'] = 9999999
    lines['category'] = 'Substation'
    lines['voltage'] = 500  # kV
    lines['trans_gids'] = '[9999]'

    # add a fake trans line for the subs to connect to to make LCP code happy
    trans_line = pd.DataFrame({'POI Name': 'fake',
                               'ac_cap': 9999999,
                               'category': 'TransLine',
                               'voltage': 500,  # kV
                               'trans_gids': None},
                              index=[9999])

    trans_line = gpd.GeoDataFrame(trans_line)
    geo = LineString([Point(0, 0), Point(100000, 100000)])
    trans_line = trans_line.set_geometry([geo], crs=crs)  # type: ignore

    pois: gpd.GeoDataFrame = pd.concat([lines, trans_line])
    pois['gid'] = pois.index

    pois.to_file(out_f, driver="GPKG")
    logger.debug('Complete')



# -*- coding: utf-8 -*-
"""
Driver class to compute exclusions
"""
import h5py
import json
import os
import logging
from abc import ABC, abstractmethod
from concurrent.futures import as_completed
from warnings import warn

import numpy as np
import geopandas as gpd
import pandas as pd
from pyproj.crs import CRS
import rasterio
from rasterio import features as rio_features
from shapely.geometry import shape

from rex import Outputs
from rex.utilities import SpawnProcessPool, log_mem
from reV.handlers.exclusions import ExclusionLayers
from reVX.handlers.geotiff import Geotiff
from reVX.handlers.layered_h5 import LayeredH5
from reVX.utilities.utilities import log_versions

logger = logging.getLogger(__name__)


class AbstractExclusionCalculatorInterface(ABC):
    """Abstract Exclusion Calculator Interface. """

    @property
    @abstractmethod
    def no_exclusions_array(self):
        """np.array: Array representing no exclusions. """
        raise NotImplementedError

    @property
    @abstractmethod
    def exclusion_merge_func(self):
        """callable: Function to merge overlapping exclusion layers. """
        raise NotImplementedError

    @abstractmethod
    def pre_process_regulations(self):
        """Reduce regulations to correct state and features.

        When implementing this method, make sure to update
        `self._regulations.df`.
        """
        raise NotImplementedError

    @abstractmethod
    def _local_exclusions_arguments(self, regulation_value, county):
        """Compile and yield arguments to `compute_local_exclusions`.

        This method should yield lists or tuples of extra args to be
        passed to `compute_local_exclusions`. Do not include the
        `regulation_value` or `county`.

        Parameters
        ----------
        regulation_value : float | int
            Regulation value for county.
        county : geopandas.GeoDataFrame
            Regulations for a single county.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def compute_local_exclusions(regulation_value, county, *args):
        """Compute local feature exclusions.

        This method should compute the exclusions using the information
        about the input county.

        Parameters
        ----------
        regulation_value : float | int
            Regulation value for county.
        county : geopandas.GeoDataFrame
            Regulations for a single county.
        *args
            Other arguments required for local exclusion calculation.

        Returns
        -------
        exclusions : np.ndarray
            Array of exclusions.
        slices : 2-tuple of `slice`
            X and Y slice objects defining where in the original array
            the exclusion data should go.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_generic_exclusions(self, max_workers=None):
        """Compute generic exclusions.

        This method should compute the exclusions using a generic
        regulation value (`self._regulations.generic`).

        Parameters
        ----------
        max_workers : int, optional
            Number of workers to use for exclusions computation, if 1
            run in serial, if > 1 run in parallel with that many
            workers, if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        exclusions : ndarray
            Raster array of exclusions
        """
        raise NotImplementedError


class AbstractBaseExclusionsMerger(AbstractExclusionCalculatorInterface):
    """
    Create exclusions layers for exclusions
    """

    def __init__(self, excl_fpath, regulations, features, hsds=False):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new exclusion layers
        regulations : `~reVX.utilities.AbstractBaseRegulations` subclass
            A regulations object used to extract exclusion regulation
            values.
        features : str
            Path to file containing features to compute exclusions from.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        """
        log_versions(logger)
        self._excl_fpath = excl_fpath
        self._regulations = regulations
        self._features = features
        self._hsds = hsds
        self._fips = self._profile = None
        self._set_profile()
        self._process_regulations(regulations.df)

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_fpath)
        return msg

    def _set_profile(self):
        """Extract profile from excl h5."""
        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as f:
            self._profile = f.profile

    def _process_regulations(self, regulations_df):
        """Parse the county regulations.

        Parse regulations, combine with county geometries from
        exclusions .h5 file. The county geometries are intersected with
        features to compute county specific exclusions.

        Parameters
        ----------
        regulations : pandas.DataFrame
            Regulations table

        Returns
        -------
        regulations: `geopandas.GeoDataFrame`
            GeoDataFrame with county level exclusion regulations merged
            with county geometries, use for intersecting with exclusion
            features.
        """
        if regulations_df is None:
            return

        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as exc:
            self._fips = exc['cnty_fips']
            cnty_fips_profile = exc.get_layer_profile('cnty_fips')

        if 'FIPS' not in regulations_df:
            msg = ('Regulations does not have county FIPS! Please add a '
                   '"FIPS" columns with the unique county FIPS values.')
            logger.error(msg)
            raise RuntimeError(msg)

        if 'geometry' not in regulations_df:
            regulations_df['geometry'] = None

        regulations_df = regulations_df[~regulations_df['FIPS'].isna()]
        regulations_df = regulations_df.set_index('FIPS')

        logger.info('Merging county geometries w/ local regulations')
        shapes_from_raster = rio_features.shapes(
            self._fips.astype(np.int32),
            transform=cnty_fips_profile['transform']
        )
        county_regs = []
        for polygon, fips_code in shapes_from_raster:
            fips_code = int(fips_code)
            if fips_code in regulations_df.index:
                local_regs = regulations_df.loc[[fips_code]].copy()
                local_regs['geometry'] = shape(polygon)
                county_regs.append(local_regs)

        if county_regs:
            regulations_df = pd.concat(county_regs)

        regulations_df = gpd.GeoDataFrame(
            regulations_df,
            crs=self.profile['crs'],
            geometry='geometry'
        )
        regulations_df = regulations_df.reset_index()
        regulations_df = regulations_df.to_crs(crs=self.profile['crs'])
        self._regulations.df = regulations_df

    @property
    def profile(self):
        """dict: Geotiff profile. """
        return self._profile

    @property
    def regulations_table(self):
        """Regulations table.

        Returns
        -------
        geopandas.GeoDataFrame | None
        """
        return self._regulations.df

    @regulations_table.setter
    def regulations_table(self, regulations_table):
        self._process_regulations(regulations_table)

    def _write_exclusions(self, geotiff, exclusions, replace=False):
        """
        Write exclusions to geotiff, replace if requested

        Parameters
        ----------
        geotiff : str
            Path to geotiff file to save exclusions too
        exclusions : ndarray
            Rasterized array of exclusions.
        replace : bool, optional
            Flag to replace local layer data with arr if file already
            exists on disk. By default `False`.
        """
        if os.path.exists(geotiff):
            _error_or_warn(geotiff, replace)

        Geotiff.write(geotiff, self.profile, exclusions)

    def _write_layer(self, out_layer, exclusions, replace=False):
        """Write exclusions to H5, replace if requested

        Parameters
        ----------
        out_layer : str
            Name of new exclusion layer to add to h5.
        exclusions : ndarray
            Rasterized array of exclusions.
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exclusion .h5 file. By default `False`.
        """
        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as exc:
            layers = exc.layers

        if out_layer in layers:
            _error_or_warn(out_layer, replace)

        try:
            description = self.description
        except AttributeError:
            description = None

        LayeredH5(self._excl_fpath).write_layer_to_h5(exclusions, out_layer,
                                                      self.profile,
                                                      description=description)

    def _county_exclusions(self):
        """Yield county exclusion arguments. """
        for ind, regulation_info in enumerate(self._regulations, start=1):
            exclusion, cnty = regulation_info
            logger.debug('Computing exclusions for {}/{} counties'
                         .format(ind, len(self.regulations_table)))
            for args in self._local_exclusions_arguments(exclusion, cnty):
                yield exclusion, cnty, args

    def compute_all_local_exclusions(self, max_workers=None):
        """Compute local exclusions for all counties either.

        Parameters
        ----------
        max_workers : int, optional
            Number of workers to use for exclusions computation, if 1
            run in serial, if > 1 run in parallel with that many
            workers, if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        exclusions : ndarray
            Raster array of exclusions.
        """
        mw = max_workers or os.cpu_count()

        log_mem(logger)
        exclusions = None
        if mw > 1:
            logger.info('Computing local exclusions in parallel using {} '
                        'workers'.format(mw))
            spp_kwargs = {"max_workers": mw, "loggers": [__name__, 'reVX']}
            with SpawnProcessPool(**spp_kwargs) as exe:
                exclusions = self._compute_local_exclusions_in_chunks(exe, mw)

        else:
            logger.info('Computing local exclusions in serial')
            for ind, cnty_inf in enumerate(self._county_exclusions(), start=1):
                exclusion_value, cnty, args = cnty_inf
                out = self.compute_local_exclusions(exclusion_value, cnty,
                                                    *args)
                local_exclusions, slices = out
                fips = cnty['FIPS'].unique()
                exclusions = self._combine_exclusions(exclusions,
                                                      local_exclusions,
                                                      fips, slices)
                logger.debug("Computed exclusions for {:,} counties"
                             .format(ind))
        if exclusions is None:
            exclusions = self.no_exclusions_array

        return exclusions

    def _compute_local_exclusions_in_chunks(self, exe, max_submissions):
        """Compute exclusions in parallel using futures. """
        futures, exclusions = {}, None

        for ind, reg in enumerate(self._county_exclusions(), start=1):
            exclusion_value, cnty, args = reg
            future = exe.submit(self.compute_local_exclusions,
                                exclusion_value, cnty, *args)
            futures[future] = cnty['FIPS'].unique()
            if ind % max_submissions == 0:
                exclusions = self._collect_local_futures(futures, exclusions)
        exclusions = self._collect_local_futures(futures, exclusions)
        return exclusions

    def _collect_local_futures(self, futures, exclusions):
        """Collect all futures from the input dictionary. """
        for future in as_completed(futures):
            new_exclusions, slices = future.result()
            exclusions = self._combine_exclusions(exclusions,
                                                  new_exclusions,
                                                  futures.pop(future),
                                                  slices=slices)
            log_mem(logger)
        return exclusions

    def compute_exclusions(self, out_layer=None, out_tiff=None, replace=False,
                           max_workers=None):
        """
        Compute exclusions for all states either in serial or parallel.
        Existing exclusions are computed if a regulations file was
        supplied during class initialization, otherwise generic exclusions
        are computed.

        Parameters
        ----------
        out_layer : str, optional
            Name to save rasterized exclusions under in .h5 file.
            If `None`, exclusions will not be written to the .h5 file.
            By default `None`.
        out_tiff : str, optional
            Path to save geotiff containing rasterized exclusions.
            If `None`, exclusions will not be written to a geotiff file.
            By default `None`.
        replace : bool, optional
            Flag to replace geotiff if it already exists.
            By default `False`.
        max_workers : int, optional
            Number of workers to use for exclusion computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None`, run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        exclusions : ndarray
            Raster array of exclusions
        """
        exclusions = self._compute_merged_exclusions(max_workers=max_workers)

        if out_layer is not None:
            logger.info('Saving exclusion layer to {} as {}'
                        .format(self._excl_fpath, out_layer))
            self._write_layer(out_layer, exclusions, replace=replace)

        if out_tiff is not None:
            logger.debug('Writing exclusions to {}'.format(out_tiff))
            self._write_exclusions(out_tiff, exclusions, replace=replace)

        return exclusions

    def _compute_merged_exclusions(self, max_workers=None):
        """Compute and merge local and generic exclusions, if necessary. """
        mw = max_workers

        if self._regulations.locals_exist:
            self.pre_process_regulations()

        generic_exclusions_exist = self._regulations.generic_exists
        local_exclusions_exist = self._regulations.locals_exist

        if not generic_exclusions_exist and not local_exclusions_exist:
            msg = ("Found no exclusions to compute: No regulations detected, "
                   "and generic multiplier not set.")
            logger.warning(msg)
            warn(msg)
            return self.no_exclusions_array

        if generic_exclusions_exist and not local_exclusions_exist:
            return self.compute_generic_exclusions(max_workers=mw)

        if local_exclusions_exist and not generic_exclusions_exist:
            local_excl = self.compute_all_local_exclusions(max_workers=mw)
            # merge ensures local exclusions are clipped county boundaries
            return self._merge_exclusions(None, local_excl)

        generic_exclusions = self.compute_generic_exclusions(max_workers=mw)
        local_exclusions = self.compute_all_local_exclusions(max_workers=mw)
        return self._merge_exclusions(generic_exclusions, local_exclusions)

    def _merge_exclusions(self, generic_exclusions, local_exclusions):
        """Merge local exclusions onto the generic exclusions."""
        logger.info('Merging local exclusions onto the generic exclusions')

        local_fips = self.regulations_table["FIPS"].unique()
        return self._combine_exclusions(generic_exclusions, local_exclusions,
                                        local_fips, replace_existing=True)

    def _combine_exclusions(self, existing, additional=None, cnty_fips=None,
                            slices=None, replace_existing=False):
        """Combine local exclusions using FIPS code"""
        if additional is None:
            return existing

        if existing is None:
            existing = self.no_exclusions_array.astype(additional.dtype)

        if slices is None:
            slices = tuple([slice(None)] * len(existing.shape))

        if cnty_fips is None:
            local_exclusions = slice(None)
        else:
            local_exclusions = np.isin(self._fips[slices], cnty_fips)

        if replace_existing:
            new_local_exclusions = additional[local_exclusions]
        else:
            new_local_exclusions = self.exclusion_merge_func(
                existing[slices][local_exclusions],
                additional[local_exclusions])
        existing[slices][local_exclusions] = new_local_exclusions
        return existing

    @classmethod
    def run(cls, excl_fpath, features_path, out_fn, regulations,
            max_workers=None, replace=False, out_layers=None, hsds=False,
            **kwargs):
        """
        Compute exclusions and write them to a geotiff. If a regulations
        file is given, compute local exclusions, otherwise compute
        generic exclusions. If both are provided, generic and local
        exclusions are merged such that the local exclusions override
        the generic ones.

        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new exclusion layers.
        features_path : str
            Path to file or directory feature shape files.
            This path can contain any pattern that can be used in the
            glob function. For example, `/path/to/features/[A]*` would
            match with all the features in the directory
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.
        out_fn : str
            Path to output geotiff where exclusion data should be
            stored.
        regulations : `~reVX.utilities.AbstractBaseRegulations` subclass
            A regulations object used to extract exclusion regulation
            distances.
        max_workers : int, optional
            Number of workers to use for exclusion computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None`, run in parallel on all available cores.
            By default `None`.
        replace : bool, optional
            Flag to replace geotiff if it already exists.
            By default `False`.
        out_layers : dict, optional
            Dictionary mapping feature file names (with extension) to
            names of layers under which exclusions should be saved in
            the `excl_fpath` .h5 file. If `None` or empty dictionary,
            no layers are saved to the h5 file. By default `None`.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        **kwargs
            Keyword args to exclusions calculator class.
        """

        out_layers = out_layers or {}
        cls_init_kwargs = {"excl_fpath": excl_fpath,
                           "regulations": regulations}
        cls_init_kwargs.update(kwargs)

        if os.path.exists(out_fn) and not replace:
            msg = ('{} already exists, exclusions will not be re-computed '
                   'unless replace=True'.format(out_fn))
            logger.error(msg)
        else:
            logger.info("Computing exclusions from {} and saving "
                        "to {}".format(features_path, out_fn))
            out_layer = out_layers.get(os.path.basename(features_path))
            exclusions = cls(excl_fpath=excl_fpath, regulations=regulations,
                             features=features_path, hsds=hsds, **kwargs)
            exclusions.compute_exclusions(out_tiff=out_fn, out_layer=out_layer,
                                          max_workers=max_workers,
                                          replace=replace)


def _error_or_warn(name, replace):
    """If replace, throw warning, otherwise throw error. """
    if not replace:
        msg = ('{} already exists. To replace it set "replace=True"'
               .format(name))
        logger.error(msg)
        raise IOError(msg)

    msg = ('{} already exists and will be replaced!'.format(name))
    logger.warning(msg)
    warn(msg)
